--[[
  ****************************************************************************
  * Original: _VirtualPlates by Saiket                                       *
  * Used for EnemyGrid and modified by Cheeta                                *
  ****************************************************************************
]]

local AddOnName, me = ...;
_VirtualPlate = me;
me.Frame = CreateFrame( "Frame", nil, WorldFrame );

me.Version = GetAddOnMetadata( AddOnName, "Version" ):match( "^([%d.]+)" );

local Plates = {};
me.Plates = Plates;
local PlatesVisible = {};
me.PlatesVisible = PlatesVisible;


local PlateOverrides = {}; -- [ MethodName ] = Function overrides for Visuals

-- Individual plate methods
do
	--- If an anchor ataches to the original plate (by WoW), re-anchor to the Visual.
	local function ResetPoint ( Plate, Region, Point, RelFrame, ... )
		if ( RelFrame == Plate ) then
			Region:SetPoint( Point, Plates[ Plate ], ... );
		end
	end
	--- Re-anchors regions when a plate is shown.
	-- WoW re-anchors most regions when it shows a nameplate, so restore those anchors to the Visual frame.
	function me:PlateOnShow ()
		NextUpdate = 0; -- Resize instantly
		local Visual = Plates[ self ];
		PlatesVisible[ self ] = Visual;
		Visual:Show();

		-- Reposition all regions
		for Index, Region in ipairs( self ) do
			for Point = 1, Region:GetNumPoints() do
				ResetPoint( self, Region, Region:GetPoint( Point ) );
			end
		end
	end
end
--- Removes the plate from the visible list when hidden.
function me:PlateOnHide ()
	PlatesVisible[ self ] = nil;
	Plates[ self ]:Hide(); -- Explicitly hide so IsShown returns false.
end




-- Main plate handling and updating
do
	local WorldFrameGetChildren = WorldFrame.GetChildren;
	local select = select;
	do
		--- Parents all plate children to the Visual, and saves references to them in the plate.
		-- @param Plate  Original nameplate children are being removed from.
		-- @param ...  Children of Plate to be reparented.
		local function ReparentChildren ( Plate, ... )
			local Visual = Plates[ Plate ];
			for Index = 1, select( "#", ... ) do
				local Child = select( Index, ... );
				if ( Child ~= Visual ) then
					local LevelOffset = Child:GetFrameLevel() - Plate:GetFrameLevel();
					Child:SetParent( Visual );
					Child:SetFrameLevel( Visual:GetFrameLevel() + LevelOffset ); -- Maintain relative frame levels
					Plate[ #Plate + 1 ] = Child;
				end
			end
		end
		--- Parents all plate regions to the Visual, similar to ReparentChildren.
		-- @see ReparentChildren
		local function ReparentRegions ( Plate, ... )
			local Visual = Plates[ Plate ];
			for Index = 1, select( "#", ... ) do
				local Region = select( Index, ... );
				Region:SetParent( Visual );
				Plate[ #Plate + 1 ] = Region;
			end
		end

		--- Adds and skins a new nameplate.
		-- @param Plate  Newly found default nameplate to be hooked.
		local function PlateAdd ( Plate )
			local Visual = CreateFrame( "Frame", nil, Plate );
			Plates[ Plate ] = Visual;

			Visual:Hide(); -- Gets explicitly shown on plate show
			Visual:SetPoint( "TOPRIGHT" );
			Visual:SetSize(145, 36); --just in case

			ReparentChildren( Plate, Plate:GetChildren() );
			ReparentRegions( Plate, Plate:GetRegions() );
			Visual:EnableDrawLayer( "HIGHLIGHT" ); -- Allows the highlight to show without enabling mouse events

			Plate:SetClampedToScreen()
			Plate:SetScript( "OnShow", me.PlateOnShow );
			Plate:SetScript( "OnHide", me.PlateOnHide );
			if ( Plate:IsVisible() ) then
				me.PlateOnShow( Plate );
			end

			-- Hook methods
			for Key, Value in pairs( PlateOverrides ) do
				Visual[ Key ] = Value;
			end
		end

		--- Scans children of WorldFrame and handles new nameplates.
		-- @param ...  Children of the WorldFrame.
		local function PlatesScan ( ... )
			for Index = 1, select( "#", ... ) do
				local Frame = select( Index, ... );
				if ( not Plates[ Frame ] ) then
					local Region = Frame:GetRegions();
					if ( Region and Region:GetObjectType() == "Texture" and Region:GetTexture() == [[Interface\TargetingFrame\UI-TargetingFrame-Flash]] ) then
						PlateAdd( Frame );
					end
				end
			end
		end

		local ChildCount, NewChildCount = 0;
		--- Adds new nameplates
		function me:WorldFrameOnUpdate ( Elapsed )
			-- Check for new nameplates
			NewChildCount = self:GetNumChildren();
			if ( ChildCount ~= NewChildCount ) then
				ChildCount = NewChildCount;
				PlatesScan( WorldFrameGetChildren( self ) );
			end
		end
	end

	local unpack = unpack;
	local Children = {};
	--- Filters the results of WorldFrame:GetChildren to replace plates with their visuals.
	local function ReplaceChildren ( ... )
		local Count = select( "#", ... );
		for Index = 1, Count do
			local Frame = select( Index, ... );
			Children[ Index ] = Plates[ Frame ] or Frame;
		end
		for Index = Count + 1, #Children do -- Remove any extras from the last call
			Children[ Index ] = nil;
		end
		return unpack( Children );
	end
	--- Returns Visual frames in place of real nameplates.
	-- @return The results of WorldFrame:GetChildren with any reference to a plate replaced with its visual.
	function WorldFrame:OldGetChildren( ... )
		return WorldFrameGetChildren( self, ... )
        end
	function WorldFrame:GetChildren ( ... )
		return ReplaceChildren( WorldFrameGetChildren( self, ... ) );
	end
end




--- Initializes settings once loaded.
function me.Frame:ADDON_LOADED ( Event, AddOn )
	if ( AddOn == AddOnName ) then
		self:UnregisterEvent( Event );
		self[ Event ] = nil;
	end
end

--- Global event handler.
function me.Frame:OnEvent ( Event, ... )
	if ( self[ Event ] ) then
		return self[ Event ]( self, Event, ... );
	end
end

WorldFrame:HookScript( "OnUpdate", me.WorldFrameOnUpdate ); -- First OnUpdate handler to run
me.Frame:SetScript( "OnEvent", me.Frame.OnEvent );
me.Frame:RegisterEvent( "ADDON_LOADED" );


local GetParent = me.Frame.GetParent;
do
	--- Add method overrides to be applied to plates' Visuals.
	local function AddPlateOverride ( MethodName )
		PlateOverrides[ MethodName ] = function ( self, ... )
			local Plate = GetParent( self );
			return Plate[ MethodName ]( Plate, ... );
		end
	end
	AddPlateOverride( "GetParent" );
	AddPlateOverride( "SetAlpha" );
	AddPlateOverride( "GetAlpha" );
	AddPlateOverride( "GetEffectiveAlpha" );
end
-- Method overrides to use plates' OnUpdate script handlers instead of their Visuals' to preserve handler execution order
do
	--- Wrapper for plate OnUpdate scripts to replace their self parameter with the plate's Visual.
	local function OnUpdateOverride ( self, ... )
		self.OnUpdate( Plates[ self ], ... );
	end
	local type = type;

	local SetScript = me.Frame.SetScript;
	--- Redirects all SetScript calls for the OnUpdate handler to the original plate.
	function PlateOverrides:SetScript ( Script, Handler, ... )
		if ( type( Script ) == "string" and Script:lower() == "onupdate" ) then
			local Plate = GetParent( self );
			Plate.OnUpdate = Handler;
			return Plate:SetScript( Script, Handler and OnUpdateOverride or nil, ... );
		else
			return SetScript( self, Script, Handler, ... );
		end
	end

	local GetScript = me.Frame.GetScript;
	--- Redirects calls to GetScript for the OnUpdate handler to the original plate's script.
	function PlateOverrides:GetScript ( Script, ... )
		if ( type( Script ) == "string" and Script:lower() == "onupdate" ) then
			return GetParent( self ).OnUpdate;
		else
			return GetScript( self, Script, ... );
		end
	end

	local HookScript = me.Frame.HookScript;
	--- Redirects all HookScript calls for the OnUpdate handler to the original plate.
	-- Also passes the visual to the hook script instead of the plate.
	function PlateOverrides:HookScript ( Script, Handler, ... )
		if ( type( Script ) == "string" and Script:lower() == "onupdate" ) then
			local Plate = GetParent( self );
			if ( Plate.OnUpdate ) then
				-- Hook old OnUpdate handler
				local Backup = Plate.OnUpdate;
				function Plate:OnUpdate ( ... )
					Backup( self, ... ); -- Technically we should return Backup's results to match HookScript's hook behavior,
					return Handler( self, ... ); -- but the overhead isn't worth it when these results get discarded.
				end
			else
				Plate.OnUpdate = Handler;
			end
			return Plate:SetScript( Script, OnUpdateOverride, ... );
		else
			return HookScript( self, Script, Handler, ... );
		end
	end
end